//
//  My3DMixer.m
//  3DMixer
//
//  Created by Toshi Nagata on 08/02/02.
/*
    Copyright (c) 2008 Toshi Nagata. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "My3DMixer.h"
#import "ProgressPanelController.h"

@implementation My3DMixer

#define kOutOfMemoryError 25

void
CheckOSStatus(int status)
{
	if (status != 0) {
		NSError *err = [NSError errorWithDomain: NSOSStatusErrorDomain code: status userInfo: nil];
		NSString *msg = [err localizedDescription];
		[NSException raise: NSGenericException format: @"%@", msg];
	}
}

#pragma mark ====== Read/write audio files ======

- (void)disposeAudioFileAtIndex: (int)index
{
	if (audioFiles != NULL && index >= 0 && index <= busCount) {
		if (audioFiles[index] != NULL) {
			ExtAudioFileDispose(audioFiles[index]);
			audioFiles[index] = NULL;
		}
	}
}

- (void)disposeAudioFiles
{
	if (audioFiles != NULL) {
		int i;
		for (i = 0; i <= busCount; i++)
			[self disposeAudioFileAtIndex: i];
		free(audioFiles);
		audioFiles = NULL;
	}
}

- (void)setAudioFormat: (const AudioStreamBasicDescription *)inAudioFormat
{
	/*  Output format = given format  */
	writeFormat = *inAudioFormat;
	
	/*  Internal format = 'canonical' format with given bitrate  */
	mixerFormat = writeFormat;
	mixerFormat.mFormatID = kAudioFormatLinearPCM;
	mixerFormat.mFormatFlags = kAudioFormatFlagIsFloat | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;
	if (mixerFormat.mFormatID == kAudioFormatLinearPCM && writeFormat.mChannelsPerFrame == 1)
		mixerFormat.mFormatFlags &= ~kLinearPCMFormatFlagIsNonInterleaved;
#if __BIG_ENDIAN__
	mixerFormat.mFormatFlags |= kAudioFormatFlagIsBigEndian;
#endif
	mixerFormat.mBitsPerChannel = sizeof(Float32) * 8;
	mixerFormat.mBytesPerFrame = mixerFormat.mBitsPerChannel / 8;
	mixerFormat.mFramesPerPacket = 1;
	mixerFormat.mBytesPerPacket = mixerFormat.mBytesPerFrame;

	CheckOSStatus(AudioUnitSetProperty(mixerUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &mixerFormat, sizeof(AudioStreamBasicDescription)));
}

- (void)releaseAudioBufferAtIndex: (int)index
{
	if (bufferLists != NULL && index <= busCount) {
		int i;
		AudioBufferList *list = bufferLists[index];
		if (list != NULL) {
			for (i = 0; i < list->mNumberBuffers; i++) {
				if (list->mBuffers[i].mData != NULL)
					free(list->mBuffers[i].mData);
			}
			free(list);
			bufferLists[index] = NULL;
		}
	}
}

- (void)releaseAudioBuffers
{
	if (bufferLists != NULL) {
		int i;
		for (i = 0; i <= busCount; i++) {
			[self releaseAudioBufferAtIndex: i];
		}
		free(bufferLists);
		bufferLists = NULL;
	}
}

- (void)allocateAudioBufferAtIndex: (int)index
{
	UInt32 param, nsamples;
	int i;
	AudioBufferList *list;

	if (bufferLists == NULL) {
		bufferLists = (AudioBufferList **)calloc(sizeof(AudioBufferList *), busCount + 1);
	}
	[self releaseAudioBufferAtIndex: index];

	/*  The number of samples in the IO buffer  */
	nsamples = 512;
	
	param = mixerFormat.mChannelsPerFrame;
	list = (AudioBufferList *)calloc(1, sizeof(AudioBufferList) + param * sizeof(AudioBuffer));
	if (list == NULL)
		[NSException raise: NSMallocException format: @"Cannot allocate audio buffer"];
	list->mNumberBuffers = param;

	for (i = 0; i < param; i++) {
		list->mBuffers[i].mNumberChannels = 1;
		list->mBuffers[i].mDataByteSize = nsamples * mixerFormat.mBytesPerFrame;
		list->mBuffers[i].mData = calloc(mixerFormat.mBytesPerFrame, nsamples);
		if (list->mBuffers[i].mData == NULL)
			[NSException raise: NSMallocException format: @"Cannot allocate audio buffer"];
	}
	bufferLists[index] = list;
}

static OSStatus
readFileCallback(void* inRefCon, AudioUnitRenderActionFlags* ioActionFlags, const AudioTimeStamp* inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList* ioData)
{
	My3DMixer *mixer = (My3DMixer *)inRefCon;
	OSStatus err = noErr;
	UInt32 numberFrames;
	
	/*  Read from file  */
	numberFrames = inNumberFrames;
//	err = ExtAudioFileRead(mixer->audioFiles[inBusNumber + 1], &numberFrames, mixer->bufferLists[inBusNumber + 1]);
	err = ExtAudioFileRead(mixer->audioFiles[inBusNumber + 1], &numberFrames, ioData);
	if (numberFrames == 0) {
		/*  End of file encountered  */
		mixer->exhaustFlags[inBusNumber + 1] = YES;
	}
	if (err != noErr) {
	//	fprintf(stderr, "ExtAudioFileRead() failed with error %d\n", err);
		return err;
	}

	/*  Render into audio buffer  */
//	err = AudioUnitRender(mixer->mixerUnit, ioActionFlags, inTimeStamp, inBusNumber, inNumberFrames, mixer->bufferLists[inBusNumber + 1]);
//	if (err != noErr) {
//		fprintf(stderr, "AudioUnitRender() failed with error %d\n", err);
//		return err;
//	}
	
	return noErr;
}

static OSStatus
writeFileCallback(void* inRefCon, AudioUnitRenderActionFlags* ioActionFlags, const AudioTimeStamp* inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList* ioData)
{
	My3DMixer *mixer = (My3DMixer *)inRefCon;
	OSStatus err = noErr;

	/*  Render into audio buffer  */
	err = AudioUnitRender(mixer->mixerUnit, ioActionFlags, inTimeStamp, inBusNumber, inNumberFrames, mixer->bufferLists[0]);
	if (err != noErr) {
		fprintf(stderr, "AudioUnitRender() failed with error %d\n", err);
		return err;
	}
	
	/*  Write to file  */
	err = ExtAudioFileWriteAsync(mixer->audioFiles[0], inNumberFrames, mixer->bufferLists[0]);
	if (err != noErr) {
		fprintf(stderr, "ExtAudioFileWrite() failed with error %d\n", err);
		return err;
	}

	return noErr;
}

- (void)setInputFile: (NSString *)filename forBusIndex: (int)index
{
	ExtAudioFileRef exref;
	FSRef fref;
	AURenderCallbackStruct callback;
	
	if (audioFiles == NULL) {
		audioFiles = (ExtAudioFileRef *)calloc(sizeof(ExtAudioFileRef), busCount + 1);
	}
	[self disposeAudioFileAtIndex: index + 1];

	[self allocateAudioBufferAtIndex: index + 1];

	/*  Open an audio file  */
	CheckOSStatus(FSPathMakeRef((unsigned char *)[filename fileSystemRepresentation], &fref, NULL));
	CheckOSStatus(ExtAudioFileOpen(&fref, &exref));
	
	/*  Set the client data format (= mixer format)  */
	CheckOSStatus(ExtAudioFileSetProperty(exref, kExtAudioFileProperty_ClientDataFormat, sizeof(AudioStreamBasicDescription), &mixerFormat));

	/*  Set callback function to the mixer unit input  */
	callback.inputProc = readFileCallback;
	callback.inputProcRefCon = self;
	CheckOSStatus(AudioUnitSetProperty(mixerUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input, index, &callback, sizeof(AURenderCallbackStruct)));
	
	audioFiles[index + 1] = exref;
}

- (void)setOutputFile: (NSString *)filename
{
	ExtAudioFileRef exref;
	FSRef dirref;
	NSString *dirname, *basename;

	if (audioFiles == NULL) {
		audioFiles = (ExtAudioFileRef *)calloc(sizeof(ExtAudioFileRef), busCount + 1);
	}
	[self disposeAudioFileAtIndex: 0];
	[self allocateAudioBufferAtIndex: 0];
	
	/*  Prepare FSRefs for the parent directory and the file to create  */
	dirname = [filename stringByDeletingLastPathComponent];
	basename = [filename lastPathComponent];
	CheckOSStatus(FSPathMakeRef((unsigned char *)[dirname fileSystemRepresentation], &dirref, NULL));

	/*  Create an audio file  */
	CheckOSStatus(ExtAudioFileCreateNew(&dirref, (CFStringRef)basename, kAudioFileAIFFType, &writeFormat, NULL, &exref));
	
	/*  Set the client data format (= mixer format)  */
	CheckOSStatus(ExtAudioFileSetProperty(exref, kExtAudioFileProperty_ClientDataFormat, sizeof(AudioStreamBasicDescription), &mixerFormat));
	
	audioFiles[0] = exref;
}

#pragma mark ====== Init/dealloc ======

- (id)init
{
	self = [super init];
	return self;
}

- (void)dealloc
{
	[self releaseAudioBuffers];
	[self disposeAudioFiles];
	[super dealloc];
}

#pragma mark ====== Set up 3DMixer AU ======

- (void)create3DMixer
{
	ComponentDescription cd;
	Component comp;

    cd.componentFlags = 0;
    cd.componentFlagsMask = 0;
    cd.componentType = kAudioUnitType_Mixer;
    cd.componentSubType = kAudioUnitSubType_3DMixer;
    cd.componentManufacturer = kAudioUnitManufacturer_Apple;
	if ((comp = FindNextComponent(NULL, &cd)) != NULL) {
		CheckOSStatus(OpenAComponent(comp, &(mixerUnit)));
	}
}

- (void)setMixerBusCount: (int)count
{
	UInt32 param;
	[self disposeAudioFiles];
	[self releaseAudioBuffers];
	if (exhaustFlags != NULL)
		free(exhaustFlags);
	exhaustFlags = (BOOL *)calloc(sizeof(BOOL), count + 1);
	if (exhaustFlags == NULL)
		CheckOSStatus(kOutOfMemoryError);
	param = busCount = count;
	CheckOSStatus(AudioUnitSetProperty(mixerUnit, kAudioUnitProperty_BusCount, kAudioUnitScope_Input, 0, &param, sizeof(param)));
}

- (void)setUseInternalReverb: (BOOL)flag
{
	UInt32 useReverb = flag;
	CheckOSStatus(AudioUnitSetProperty(mixerUnit, kAudioUnitProperty_UsesInternalReverb, kAudioUnitScope_Global, 0, &useReverb, sizeof(useReverb)));
}

- (void)enableInternalReverb: (BOOL)flag forBusIndex: (int)index
{
	UInt32 render3DFlags;
	UInt32 size;
	size = sizeof(UInt32);
	CheckOSStatus(AudioUnitGetProperty(mixerUnit, kAudioUnitProperty_3DMixerRenderingFlags, kAudioUnitScope_Input, index, &render3DFlags, &size));
	if (flag)
		render3DFlags |= k3DMixerRenderingFlags_DistanceDiffusion;
	else
		render3DFlags &= ~k3DMixerRenderingFlags_DistanceDiffusion;
    CheckOSStatus(AudioUnitSetProperty(mixerUnit, kAudioUnitProperty_3DMixerRenderingFlags, kAudioUnitScope_Input, index, &render3DFlags, sizeof(render3DFlags)));
}

- (void)setAzimuth: (float)azimuth elevation: (float)elevation distance: (float)distance forBusIndex: (int)index
{
    CheckOSStatus(AudioUnitSetParameter(mixerUnit, k3DMixerParam_Azimuth, kAudioUnitScope_Input, index, azimuth, 0));
    CheckOSStatus(AudioUnitSetParameter(mixerUnit, k3DMixerParam_Elevation, kAudioUnitScope_Input, index, elevation, 0));
    CheckOSStatus(AudioUnitSetParameter(mixerUnit, k3DMixerParam_Distance, kAudioUnitScope_Input, index, distance, 0));
}

- (void)setPlaybackRate: (float)pitch forBusIndex: (int)index
{
    CheckOSStatus(AudioUnitSetParameter(mixerUnit, k3DMixerParam_PlaybackRate, kAudioUnitScope_Input, index, pitch, 0));
}

- (void)setGain: (float)gain forBusIndex: (int)index
{
    CheckOSStatus(AudioUnitSetParameter(mixerUnit, k3DMixerParam_Gain, kAudioUnitScope_Input, index, gain, 0));
}

- (void)setReferenceDistance: (float)refDist maxDistance: (float)maxDist maxAttenuation: (float)maxAttn forBusIndex: (int)index
{
	MixerDistanceParams distanceParams;
	distanceParams.mReferenceDistance = refDist;
	distanceParams.mMaxDistance = maxDist;
	distanceParams.mMaxAttenuation = maxAttn;
    CheckOSStatus(AudioUnitSetProperty(mixerUnit, kAudioUnitProperty_3DMixerDistanceParams, kAudioUnitScope_Input, index, &distanceParams, sizeof(distanceParams)));
}

- (void)setSpatializationAlgorithm: (int)algorithm forBusIndex: (int)index
{
	UInt32 myAlgorithm = algorithm;
	CheckOSStatus(AudioUnitSetProperty(mixerUnit, kAudioUnitProperty_SpatializationAlgorithm, kAudioUnitScope_Input, index, &myAlgorithm, sizeof(myAlgorithm)));
}

- (BOOL)render
{
	AudioTimeStamp tStamp;
	int i, count, interval;
	UInt32 numFrames = 512;
	SInt64 maxFileFrames = 0;
	id progressPanelController;
	BOOL retval;

	memset(&tStamp, 0, sizeof(AudioTimeStamp));
	tStamp.mFlags = kAudioTimeStampSampleTimeValid;
	count = 0;
	interval = (int)(mixerFormat.mSampleRate / numFrames);
	CheckOSStatus(AudioUnitInitialize(mixerUnit));
	memset(exhaustFlags, 0, sizeof(exhaustFlags[0]) * (busCount + 1));

	/*  Calculate the file length (in frames)  */
	for (i = 0; i < busCount; i++) {
		SInt64 fileFrames;
		UInt32 propSize = sizeof(SInt64);
		CheckOSStatus(ExtAudioFileGetProperty(audioFiles[i + 1], kExtAudioFileProperty_FileLengthFrames, &propSize, &fileFrames));
		if (maxFileFrames < fileFrames)
			maxFileFrames = fileFrames;
	}
	
	progressPanelController = [[ProgressPanelController alloc] initWithTitle: @"Processing" andCaption: @"Processing..."];
	[progressPanelController beginSession];

	retval = NO;
	while (1) {
		AudioUnitRenderActionFlags actionFlags = 0;
		CheckOSStatus(AudioUnitRender(mixerUnit, &actionFlags, &tStamp, 0, numFrames, bufferLists[0]));
		tStamp.mSampleTime += numFrames;
		CheckOSStatus(ExtAudioFileWrite(audioFiles[0], numFrames, bufferLists[0]));
		for (i = 1; i <= busCount; i++) {
			if (exhaustFlags[i] == NO)
				break;
		}
		if (i > busCount) {
			retval = YES;
			break;
		}
		if (++count % interval == 0) {
			double progress = tStamp.mSampleTime / (double)maxFileFrames;
			[progressPanelController setProgressAmount: progress * 100.0];
			if (![progressPanelController runSession] || [progressPanelController canceled])
				break;
		//	printf("Frame %ld, duration %.3f\n", (long)tStamp.mSampleTime, (float)tStamp.mSampleTime / mixerFormat.mSampleRate);
		}
	}
	[progressPanelController endSession];
	[progressPanelController close];
	[progressPanelController release];
	return retval;
}

#pragma mark ====== External entry point ======

static int
spatializationAlgorithm(int value)
{
	switch (value) {
		case 0: return kSpatializationAlgorithm_StereoPassThrough;
		case 1: return kSpatializationAlgorithm_EqualPowerPanning;
		case 2: return kSpatializationAlgorithm_HRTF;
		case 3: return kSpatializationAlgorithm_SphericalHead;
		case 4: return kSpatializationAlgorithm_VectorBasedPanning;
		case 5: return kSpatializationAlgorithm_SoundField;
		default: return kSpatializationAlgorithm_EqualPowerPanning;
	}
}

- (BOOL)renderToFile: (NSString *)filename settings: (NSDictionary *)settings busSettings: (NSArray *)busSettings
{
	@try
	{
		BOOL retval;
		int i, count;
		int algorithm;
		AudioStreamBasicDescription fmt;
		float refDistance, maxDistance, maxAttenuation;
		NSFileManager *fman = [NSFileManager defaultManager];
		NSString *tempfile;

		count = [busSettings count];
		[self create3DMixer];
		[self setMixerBusCount: count];
		[self setUseInternalReverb: YES];
		fmt.mSampleRate = [[settings valueForKey: @"bitrate"] floatValue];
		fmt.mFormatID = kAudioFormatLinearPCM;
		fmt.mFormatFlags = (kAudioFormatFlagIsBigEndian | kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked);
		fmt.mBytesPerPacket = 6;
		fmt.mFramesPerPacket = 1;
		fmt.mBytesPerFrame = 6;
		fmt.mBitsPerChannel = 24;
		fmt.mReserved = 0;
		switch ([[settings valueForKey: @"channels"] intValue]) {
			case 0: fmt.mChannelsPerFrame = 2; break;
			default: fmt.mChannelsPerFrame = 2; break;
		}
		[self setAudioFormat: &fmt];

		algorithm = spatializationAlgorithm([[settings valueForKey: @"algorithm"] intValue]);
		refDistance = [[settings valueForKey: @"reference_distance"] floatValue];
		maxDistance = [[settings valueForKey: @"max_distance"] floatValue];
		maxAttenuation = [[settings valueForKey: @"max_attenuation"] floatValue];
		for (i = 0; i < count; i++) {
			id busSetting = [busSettings objectAtIndex: i];
			[self enableInternalReverb: [[busSetting valueForKey: @"reverb"] boolValue] forBusIndex: i];
			[self setAzimuth: [[busSetting valueForKey: @"azimuth"] floatValue] elevation: [[busSetting valueForKey: @"elevation"] floatValue] distance: [[busSetting valueForKey: @"distance"] floatValue] forBusIndex: i];
			[self setSpatializationAlgorithm: algorithm forBusIndex: i];
			[self setPlaybackRate: [[busSetting valueForKey: @"pitch"] floatValue] forBusIndex: i];
			[self setGain: [[busSetting valueForKey: @"gain"] floatValue] forBusIndex: i];
			[self setReferenceDistance: refDistance maxDistance: maxDistance maxAttenuation: maxAttenuation forBusIndex: i];
			[self setInputFile: [[busSetting valueForKey: @"source_dir"] stringByAppendingPathComponent: [busSetting valueForKey: @"source"]] forBusIndex: i];
		}

		if ([fman fileExistsAtPath: filename]) {
			/*  Create a temporary file  */
			int suffix = 0;
			do {
				tempfile = [NSString stringWithFormat: @"%@.%d", filename, suffix++];
			} while ([fman fileExistsAtPath: tempfile]);
		} else tempfile = filename;
		[self setOutputFile: tempfile];
		if (![self render]) {
			NSRunAlertPanel(@"Error", @"Creating a new audio file failed.", nil, nil, nil);
			retval = NO;
		} else retval = YES;
		[self disposeAudioFiles];
		[self releaseAudioBuffers];
		if (retval) {
			if (tempfile != filename) {
				[fman removeFileAtPath: filename handler: nil];
				[fman movePath: tempfile toPath: filename handler: nil];
			}
		} else {
			[fman removeFileAtPath: tempfile handler: nil];
		}
		return retval;
	}
	@catch (NSException *exp)
	{
		NSRunAlertPanel(@"Error", [exp reason], nil, nil, nil);
		return NO;
	}
	return YES;
}

@end
